@@ -5,16 +5,17 @@ from typing import Union, Tuple
 import pydantic
 import numpy as np
 
-from .types import Ax, EMField, ArrayLike, Bound, FreqArray
-from .types import Literal, Direction, Coordinate, Axis, ObsGridArray, RadVec
+from .types import Ax, EMField, ArrayLike, FreqArray
+from .types import Literal, Direction, Coordinate, Axis, ObsGridArray
 from .geometry import Box
-from .medium import Medium
-from .validators import assert_plane, validate_unique
-from .base import cached_property
+from .validators import assert_plane
+from .base import cached_property, Tidy3dBaseModel
 from .mode import ModeSpec
+from .apodization import ApodizationSpec
 from .viz import PlotParams, plot_params_monitor, ARROW_COLOR_MONITOR, ARROW_ALPHA
-from ..log import SetupError, log, DataError, ValidationError
 from ..constants import HERTZ, SECOND, MICROMETER, RADIAN, inf
+from ..exceptions import SetupError, ValidationError
+from ..log import log
 
 
 BYTES_REAL = 4
@@ -75,6 +76,16 @@ class FreqMonitor(Monitor, ABC):
         units=HERTZ,
     )
 
+    apodization: ApodizationSpec = pydantic.Field(
+        ApodizationSpec(),
+        title="Apodization Specification",
+        description="Sets parameters of (optional) apodization. Apodization applies a windowing "
+        "function to the Fourier transform of the time-domain fields into frequency-domain ones, "
+        "and can be used to truncate the beginning and/or end of the time signal, for example "
+        "to eliminate the source pulse when studying the eigenmodes of a system. Note: apodization "
+        "affects the normalization of the frequency-domain fields.",
+    )
+
     @pydantic.validator("freqs", always=True)
     def _freqs_non_empty(cls, val):
         """Assert one frequency present."""
@@ -118,6 +129,7 @@ class TimeMonitor(Monitor, ABC):
     def time_inds(self, tmesh: ArrayLike[float, 1]) -> Tuple[int, int]:
         """Compute the starting and stopping index of the monitor in a given discrete time mesh."""
 
+        tmesh = np.array(tmesh)
         tind_beg, tind_end = (0, 0)
 
         if tmesh.size == 0:
@@ -189,7 +201,7 @@ class AbstractFieldMonitor(Monitor, ABC):
         """Given a tuple of the number of cells spanned by the monitor along each dimension,
         return the number of cells one would have after downsampling based on ``interval_space``.
         """
-        num_cells_new = list(num_cells.copy())
+        num_cells_new = list(num_cells)
         for idx, interval in enumerate(self.interval_space):
             if interval == 1 or num_cells[idx] < 4 or (num_cells[idx] - 1) <= interval:
                 continue
@@ -203,6 +215,11 @@ class PlanarMonitor(Monitor, ABC):
 
     _plane_validator = assert_plane()
 
+    @cached_property
+    def normal_axis(self) -> Axis:
+        """Axis normal to the monitor's plane."""
+        return self.size.index(0.0)
+
 
 class AbstractModeMonitor(PlanarMonitor, FreqMonitor):
     """:class:`Monitor` that records mode-related data."""
@@ -219,10 +236,8 @@ class AbstractModeMonitor(PlanarMonitor, FreqMonitor):
         y: float = None,
         z: float = None,
         ax: Ax = None,
-        sim_bounds: Bound = None,
         **patch_kwargs,
     ) -> Ax:
-
         # call the monitor.plot() function first
         ax = super().plot(x=x, y=y, z=z, ax=ax, **patch_kwargs)
 
@@ -236,10 +251,11 @@ class AbstractModeMonitor(PlanarMonitor, FreqMonitor):
             z=z,
             ax=ax,
             direction=self._dir_arrow,
+            bend_radius=self.mode_spec.bend_radius,
+            bend_axis=self._bend_axis,
             color=ARROW_COLOR_MONITOR,
             alpha=arrow_alpha,
             both_dirs=True,
-            sim_bounds=sim_bounds,
         )
         return ax
 
@@ -249,7 +265,16 @@ class AbstractModeMonitor(PlanarMonitor, FreqMonitor):
         dx = np.cos(self.mode_spec.angle_phi) * np.sin(self.mode_spec.angle_theta)
         dy = np.sin(self.mode_spec.angle_phi) * np.sin(self.mode_spec.angle_theta)
         dz = np.cos(self.mode_spec.angle_theta)
-        return self.unpop_axis(dz, (dx, dy), axis=self.size.index(0.0))
+        return self.unpop_axis(dz, (dx, dy), axis=self.normal_axis)
+
+    @cached_property
+    def _bend_axis(self) -> Axis:
+        if self.mode_spec.bend_radius is None:
+            return None
+        in_plane = [0, 0]
+        in_plane[self.mode_spec.bend_axis] = 1
+        direction = self.unpop_axis(0, in_plane, axis=self.normal_axis)
+        return direction.index(1)
 
 
 class FieldMonitor(AbstractFieldMonitor, FreqMonitor):
@@ -332,6 +357,18 @@ class SurfaceIntegrationMonitor(Monitor, ABC):
         description="Surfaces to exclude in the integration, if a volume monitor.",
     )
 
+    @property
+    def integration_surfaces(self):
+        """Surfaces of the monitor where fields will be recorded for subsequent integration."""
+        if self.size.count(0.0) == 0:
+            monitor_dict = self.dict()
+            exclude_surfaces = monitor_dict.pop("exclude_surfaces", None)
+            surface_monitors = self.surfaces(**monitor_dict)
+            if exclude_surfaces:
+                return [mnt for mnt in surface_monitors if mnt.name[-2:] not in exclude_surfaces]
+            return surface_monitors
+        return [self]
+
     @pydantic.root_validator(skip_on_failure=True)
     def normal_dir_exists_for_surface(cls, values):
         """If the monitor is a surface, set default ``normal_dir`` if not provided.
@@ -384,14 +421,6 @@ class FluxMonitor(AbstractFluxMonitor, FreqMonitor):
     ...     size=(2,2,0),
     ...     freqs=[200e12, 210e12],
     ...     name='flux_monitor')
-
-    Note
-    ----
-    For a 2D plane, the flux is summed up over all Yee grid pixels that are touched by the plane,
-    rather than integrated over the exact span of the plane. For a 3D monitor, this is also the
-    case, but care is taken to not over- or under-count the power at the edges. Because of this,
-    there can be small discrepancies between using a 3D FluxMonitor and manually placing six
-    2D monitors at the surface locations.
     """
 
     def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
@@ -416,14 +445,6 @@ class FluxTimeMonitor(AbstractFluxMonitor, TimeMonitor):
     ...     stop=5e-13,
     ...     interval=2,
     ...     name='flux_vs_time')
-
-    Note
-    ----
-    For a 2D plane, the flux is summed up over all Yee grid pixels that are touched by the plane,
-    rather than integrated over the exact span of the plane. For a 3D monitor, this is also the
-    case, but care is taken to not over- or under-count the power at the edges. Because of this,
-    there can be small discrepancies between using a 3D FluxMonitor and manually placing six
-    2D monitors at the surface locations.
     """
 
     def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
@@ -472,18 +493,54 @@ class ModeSolverMonitor(AbstractModeMonitor):
         """Size of monitor storage given the number of points after discretization."""
         return 6 * BYTES_COMPLEX * num_cells * len(self.freqs) * self.mode_spec.num_modes
 
+    @property
+    def interval_space(self):
+        """No downsampling is applied to the stored fields."""
+        return (1, 1, 1)
 
-class AbstractNear2FarMonitor(SurfaceIntegrationMonitor, FreqMonitor):
-    """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
-    and invokes the computation of far fields.
-    """
+    @property
+    def colocate(self):
+        """Fields are stored on the Yee grid."""
+        return False
 
-    fields: Tuple[RadVec, ...] = pydantic.Field(
-        ["Ntheta", "Nphi", "Ltheta", "Lphi"],
-        title="Field Components",
-        description="Collection of radiation vector components to store in the monitor.",
+
+class FieldProjectionSurface(Tidy3dBaseModel):
+    """Data structure to store surface monitors where near fields are recorded for
+    field projections."""
+
+    monitor: FieldMonitor = pydantic.Field(
+        ...,
+        title="Field monitor",
+        description=":class:`.FieldMonitor` on which near fields will be sampled and integrated.",
     )
 
+    normal_dir: Direction = pydantic.Field(
+        ...,
+        title="Normal vector orientation",
+        description=":class:`.Direction` of the surface monitor's normal vector w.r.t.\
+ the positive x, y or z unit vectors. Must be one of '+' or '-'.",
+    )
+
+    @cached_property
+    def axis(self) -> Axis:
+        """Returns the :class:`.Axis` normal to this surface."""
+        # assume that the monitor's axis is in the direction where the monitor is thinnest
+        return self.monitor.size.index(0.0)
+
+    @pydantic.validator("monitor", always=True)
+    def is_plane(cls, val):
+        """Ensures that the monitor is a plane, i.e., its `size` attribute has exactly 1 zero"""
+        size = val.size
+        if size.count(0.0) != 1:
+            raise ValidationError(f"Monitor '{val.name}' must be planar, given size={size}")
+        return val
+
+
+class AbstractFieldProjectionMonitor(SurfaceIntegrationMonitor, FreqMonitor):
+    """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
+    and projects them to a given set of observation points.
+    """
+
     custom_origin: Coordinate = pydantic.Field(
         None,
         title="Local origin",
@@ -492,24 +549,25 @@ class AbstractNear2FarMonitor(SurfaceIntegrationMonitor, FreqMonitor):
         units=MICROMETER,
     )
 
-    medium: Medium = pydantic.Field(
-        Medium(permittivity=1),
-        title="Background medium",
-        description="Background medium in which to radiate near fields to far fields. "
-        "If not provided, uses free space.",
+    far_field_approx: bool = pydantic.Field(
+        True,
+        title="Far field approximation",
+        description="Whether to enable the far field approximation when projecting fields.",
     )
 
     @property
-    def axis(self) -> Axis:
-        """Returns the :class:`.Axis` normal to this surface."""
-        # assume that the monitor's axis is in the direction where the monitor is thinnest
-        # and errors if the monitor is a box
-        if self.size.count(0.0) != 1:
-            raise DataError(
-                "Requested ``axis`` property for a box monitor; ``axis`` is defined "
-                "for surface monitors only."
+    def projection_surfaces(self) -> Tuple[FieldProjectionSurface, ...]:
+        """Surfaces of the monitor where near fields will be recorded for subsequent projection."""
+        surfaces = self.integration_surfaces
+        return [
+            FieldProjectionSurface(
+                monitor=FieldMonitor(
+                    center=surface.center, size=surface.size, freqs=self.freqs, name=surface.name
+                ),
+                normal_dir=surface.normal_dir,
             )
-        return self.size.index(0.0)
+            for surface in surfaces
+        ]
 
     @property
     def local_origin(self) -> Coordinate:
@@ -519,13 +577,13 @@ class AbstractNear2FarMonitor(SurfaceIntegrationMonitor, FreqMonitor):
         return self.custom_origin
 
 
-class Near2FarAngleMonitor(AbstractNear2FarMonitor):
+class FieldProjectionAngleMonitor(AbstractFieldProjectionMonitor):
     """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
-    and invokes the computation of far fields at given observation angles.
+    and projects them at given observation angles.
 
     Example
     -------
-    >>> monitor = Near2FarAngleMonitor(
+    >>> monitor = FieldProjectionAngleMonitor(
     ...     center=(1,2,3),
     ...     size=(2,2,2),
     ...     freqs=[250e12, 300e12],
@@ -536,36 +594,43 @@ class Near2FarAngleMonitor(AbstractNear2FarMonitor):
     ...     )
     """
 
+    proj_distance: float = pydantic.Field(
+        1e6,
+        title="Projection distance",
+        description="Radial distance of the projection points from ``local_origin``.",
+        units=MICROMETER,
+    )
+
     theta: ObsGridArray = pydantic.Field(
         ...,
-        title="Polar Angles",
-        description="Polar angles relative to ``local_origin`` at which to compute far fields.",
+        title="Polar angles",
+        description="Polar angles with respect to the global z axis, relative to the location of "
+        "``local_origin``, at which to project fields.",
         units=RADIAN,
     )
 
     phi: ObsGridArray = pydantic.Field(
         ...,
-        title="Azimuth Angles",
-        description="Azimuth angles relative to ``local_origin`` at which to compute far fields.",
+        title="Azimuth angles",
+        description="Azimuth angles with respect to the global z axis, relative to the location of "
+        "``local_origin``, at which to project fields.",
         units=RADIAN,
     )
 
     def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
         """Size of monitor storage given the number of points after discretization."""
         # stores 1 complex number per pair of angles, per frequency,
-        # for N_theta, N_phi, L_theta, and L_phi (4 components)
-        return BYTES_COMPLEX * len(self.theta) * len(self.phi) * len(self.freqs) * 4
+        # for Er, Etheta, Ephi, Hr, Htheta, and Hphi (6 components)
+        return BYTES_COMPLEX * len(self.theta) * len(self.phi) * len(self.freqs) * 6
 
 
-class Near2FarCartesianMonitor(AbstractNear2FarMonitor):
+class FieldProjectionCartesianMonitor(AbstractFieldProjectionMonitor):
     """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
-    and invokes the computation of far fields on a Cartesian observation plane.
-    The far-field approximation is used, so results projected to points close to
-    the monitor's position, compared to its size, may not be accurate.
+    and projects them on a Cartesian observation plane.
 
     Example
     -------
-    >>> monitor = Near2FarCartesianMonitor(
+    >>> monitor = FieldProjectionCartesianMonitor(
     ...     center=(1,2,3),
     ...     size=(2,2,2),
     ...     freqs=[250e12, 300e12],
@@ -573,103 +638,123 @@ class Near2FarCartesianMonitor(AbstractNear2FarMonitor):
     ...     custom_origin=(1,2,3),
     ...     x=[-1, 0, 1],
     ...     y=[-2, -1, 0, 1, 2],
-    ...     plane_axis=2,
-    ...     plane_distance=5
+    ...     proj_axis=2,
+    ...     proj_distance=5
     ...     )
     """
 
-    plane_axis: Axis = pydantic.Field(
+    proj_axis: Axis = pydantic.Field(
         ...,
-        title="Observation plane axis",
+        title="Projection plane axis",
         description="Axis along which the observation plane is oriented.",
     )
 
-    plane_distance: float = pydantic.Field(
-        ...,
-        title="Observation plane signed distance",
-        description="Signed distance of the observation plane along ``plane_axis`` "
-        "w.r.t. ``local_origin``",
+    proj_distance: float = pydantic.Field(
+        1e6,
+        title="Projection distance",
+        description="Signed distance of the projection plane along ``proj_axis``. "
+        "from the plane containing ``local_origin``.",
+        units=MICROMETER,
     )
 
     x: ObsGridArray = pydantic.Field(
         ...,
         title="Local x observation coordinates",
-        description="Local x observation coordinates w.r.t. ``local_origin`` and ``plane_axis``. "
-        "When ``plane_axis`` is 0, this corresponds to the global y axis. "
-        "When ``plane_axis`` is 1, this corresponds to the global x axis. "
-        "When ``plane_axis`` is 2, this corresponds to the global x axis. ",
+        description="Local x observation coordinates w.r.t. ``local_origin`` and ``proj_axis``. "
+        "When ``proj_axis`` is 0, this corresponds to the global y axis. "
+        "When ``proj_axis`` is 1, this corresponds to the global x axis. "
+        "When ``proj_axis`` is 2, this corresponds to the global x axis. ",
         units=MICROMETER,
     )
 
     y: ObsGridArray = pydantic.Field(
         ...,
         title="Local y observation coordinates",
-        description="Local y observation coordinates w.r.t. ``local_origin`` and ``plane_axis``. "
-        "When ``plane_axis`` is 0, this corresponds to the global z axis. "
-        "When ``plane_axis`` is 1, this corresponds to the global z axis. "
-        "When ``plane_axis`` is 2, this corresponds to the global y axis. ",
+        description="Local y observation coordinates w.r.t. ``local_origin`` and ``proj_axis``. "
+        "When ``proj_axis`` is 0, this corresponds to the global z axis. "
+        "When ``proj_axis`` is 1, this corresponds to the global z axis. "
+        "When ``proj_axis`` is 2, this corresponds to the global y axis. ",
         units=MICROMETER,
     )
 
     def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
         """Size of monitor storage given the number of points after discretization."""
         # stores 1 complex number per pair of grid points, per frequency,
-        # for N_theta, N_phi, L_theta, and L_phi (4 components)
-        return BYTES_COMPLEX * len(self.x) * len(self.y) * len(self.freqs) * 4
+        # for Er, Etheta, Ephi, Hr, Htheta, and Hphi (6 components)
+        return BYTES_COMPLEX * len(self.x) * len(self.y) * len(self.freqs) * 6
 
 
-class Near2FarKSpaceMonitor(AbstractNear2FarMonitor):
+class FieldProjectionKSpaceMonitor(AbstractFieldProjectionMonitor):
     """:class:`Monitor` that samples electromagnetic near fields in the frequency domain
-    and invokes the computation of far fields on an observation plane defined in k-space.
+    and projects them on an observation plane defined in k-space.
 
     Example
     -------
-    >>> monitor = Near2FarKSpaceMonitor(
+    >>> monitor = FieldProjectionKSpaceMonitor(
     ...     center=(1,2,3),
     ...     size=(2,2,2),
     ...     freqs=[250e12, 300e12],
     ...     name='n2f_monitor',
     ...     custom_origin=(1,2,3),
-    ...     u_axis=2,
-    ...     ux=[1,2],
-    ...     uy=[3,4,5]
+    ...     proj_axis=2,
+    ...     ux=[0.1,0.2],
+    ...     uy=[0.3,0.4,0.5]
     ...     )
     """
 
-    u_axis: Axis = pydantic.Field(
+    proj_axis: Axis = pydantic.Field(
         ...,
-        title="Observation plane axis",
+        title="Projection plane axis",
         description="Axis along which the observation plane is oriented.",
     )
 
+    proj_distance: float = pydantic.Field(
+        1e6,
+        title="Projection distance",
+        description="Radial distance of the projection points from ``local_origin``.",
+        units=MICROMETER,
+    )
+
     ux: ObsGridArray = pydantic.Field(
         ...,
         title="Normalized kx",
         description="Local x component of wave vectors on the observation plane, "
-        "relative to ``local_origin`` and oriented with respect to ``u_axis``, "
+        "relative to ``local_origin`` and oriented with respect to ``proj_axis``, "
         "normalized by (2*pi/lambda) where lambda is the wavelength "
-        "associated with the background medium.",
+        "associated with the background medium. Must be in the range [-1, 1].",
     )
 
     uy: ObsGridArray = pydantic.Field(
         ...,
         title="Normalized ky",
         description="Local y component of wave vectors on the observation plane, "
-        "relative to ``local_origin`` and oriented with respect to ``u_axis``, "
+        "relative to ``local_origin`` and oriented with respect to ``proj_axis``, "
         "normalized by (2*pi/lambda) where lambda is the wavelength "
-        "associated with the background medium.",
+        "associated with the background medium. Must be in the range [-1, 1].",
     )
 
+    @pydantic.root_validator()
+    def reciprocal_vector_range(cls, values):
+        """Ensure that ux, uy are in [-1, 1]."""
+        maxabs_ux = max(list(values.get("ux")), key=abs)
+        maxabs_uy = max(list(values.get("uy")), key=abs)
+        name = values.get("name")
+        if maxabs_ux > 1:
+            raise SetupError(f"Entries of 'ux' must lie in the range [-1, 1] for monitor {name}.")
+        if maxabs_uy > 1:
+            raise SetupError(f"Entries of 'uy' must lie in the range [-1, 1] for monitor {name}.")
+        return values
+
     def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
         """Size of monitor storage given the number of points after discretization."""
         # stores 1 complex number per pair of grid points, per frequency,
-        # for N_theta, N_phi, L_theta, and L_phi (4 components)
-        return BYTES_COMPLEX * len(self.ux) * len(self.uy) * len(self.freqs) * 4
+        # for Er, Etheta, Ephi, Hr, Htheta, and Hphi (6 components)
+        return BYTES_COMPLEX * len(self.ux) * len(self.uy) * len(self.freqs) * 6
 
 
 class DiffractionMonitor(PlanarMonitor, FreqMonitor):
     """:class:`Monitor` that uses a 2D Fourier transform to compute the
-    diffraction amplitudes and efficiency for given (or all) diffraction orders.
+    diffraction amplitudes and efficiency for allowed diffraction orders.
 
     Example
     -------
@@ -679,8 +764,6 @@ class DiffractionMonitor(PlanarMonitor, FreqMonitor):
     ...     freqs=[250e12, 300e12],
     ...     name='diffraction_monitor',
     ...     normal_dir='+',
-    ...     orders_x=[0, 1, 2],
-    ...     orders_y=[0],
     ...     )
     """
 
@@ -692,25 +775,6 @@ class DiffractionMonitor(PlanarMonitor, FreqMonitor):
         "Defaults to ``'+'`` if not provided.",
     )
 
-    orders_x: Tuple[int, ...] = pydantic.Field(
-        [0],
-        title="Diffraction orders along x",
-        description="Diffraction orders along x for which efficiency should be computed. "
-        "Orders corresponding to wave numbers outside of the light cone will be ignored. "
-        "By default, only order 0 will be considered.",
-    )
-
-    orders_y: Tuple[int, ...] = pydantic.Field(
-        [0],
-        title="Diffraction orders along y",
-        description="Diffraction orders along y for which efficiency should be computed. "
-        "Orders corresponding to wave numbers outside of the light cone will be ignored. "
-        "By default, only order 0 will be considered.",
-    )
-
-    _unique_x = validate_unique("orders_x")
-    _unique_y = validate_unique("orders_y")
-
     @pydantic.validator("size", always=True)
     def diffraction_monitor_size(cls, val):
         """Ensure that the monitor is infinite in the transverse direction."""
@@ -721,15 +785,10 @@ class DiffractionMonitor(PlanarMonitor, FreqMonitor):
             )
         return val
 
-    @property
-    def normal_axis(self) -> Axis:
-        """Axis normal to the monitor's plane."""
-        return self.size.index(0)
-
     def storage_size(self, num_cells: int, tmesh: ArrayLike[float, 1]) -> int:
         """Size of monitor storage given the number of points after discretization."""
-        # stores 1 complex number per pair of diffraction orders, per frequency
-        return BYTES_COMPLEX * len(self.orders_x) * len(self.orders_y) * len(self.freqs)
+        # assumes 1 diffraction order per frequency; actual size will be larger
+        return BYTES_COMPLEX * len(self.freqs)
 
 
 # types of monitors that are accepted by simulation
@@ -741,8 +800,8 @@ MonitorType = Union[
     FluxTimeMonitor,
     ModeMonitor,
     ModeSolverMonitor,
-    Near2FarAngleMonitor,
-    Near2FarCartesianMonitor,
-    Near2FarKSpaceMonitor,
+    FieldProjectionAngleMonitor,
+    FieldProjectionCartesianMonitor,
+    FieldProjectionKSpaceMonitor,
     DiffractionMonitor,
 ]
